package org.processmining.datapetrinets;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.syntax.ParseException;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithDataFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.utils.ConnectionManagerHelper;

import com.google.common.collect.ImmutableSet;

public interface DataPetriNet extends org.processmining.models.graphbased.directed.petrinet.PetrinetGraph {

	public interface PetrinetWithMarkings {
		Petrinet getNet();

		Marking getInitialMarking();

		Marking[] getFinalMarkings();
	}

	public final class Factory {

		private static final class ImmutablePetriNetAsDPN implements DataPetriNetsWithMarkings {

			private final PetrinetGraph net;
			private final Marking initialMarking;
			private final Marking finalMarking;

			private ImmutablePetriNetAsDPN(PetrinetGraph net) {
				this(net, null, null);
			}

			public ImmutablePetriNetAsDPN(PetrinetGraph net, Marking initialMarking, Marking finalMarking) {
				this.net = net;
				this.initialMarking = initialMarking;
				this.finalMarking = finalMarking;
				
			}

			public int compareTo(
					DirectedGraph<PetrinetNode, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> o) {
				return net.compareTo(o);
			}

			public AttributeMap getAttributeMap() {
				return net.getAttributeMap();
			}

			public DirectedGraph<?, ?> getGraph() {
				return net.getGraph();
			}

			public void removeNode(DirectedGraphNode cell) {
				throw new UnsupportedOperationException("Immutable");
			}

			@SuppressWarnings("rawtypes") // non generic in supertype
			public void removeEdge(DirectedGraphEdge edge) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> getOutEdges(
					DirectedGraphNode node) {
				return net.getOutEdges(node);
			}

			public Set<PetrinetNode> getNodes() {
				return net.getNodes();
			}

			public Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> getInEdges(
					DirectedGraphNode node) {
				return net.getInEdges(node);
			}

			public Set<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> getEdges() {
				return net.getEdges();
			}

			public Transition removeTransition(Transition transition) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Place removePlace(Place place) {
				throw new UnsupportedOperationException("Immutable");
			}

			public ExpandableSubNet removeGroup(ExpandableSubNet transition) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Arc removeArc(PetrinetNode source, PetrinetNode target) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Collection<Transition> getTransitions() {
				return net.getTransitions();
			}

			public Collection<Place> getPlaces() {
				return net.getPlaces();
			}

			public String getLabel() {
				return net.getLabel();
			}

			public Collection<ExpandableSubNet> getGroups() {
				return net.getGroups();
			}

			public Arc getArc(PetrinetNode source, PetrinetNode target) {
				return net.getArc(source, target);
			}

			public Transition addTransition(String label, ExpandableSubNet parent) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Transition addTransition(String label) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Place addPlace(String label, ExpandableSubNet parent) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Place addPlace(String label) {
				throw new UnsupportedOperationException("Immutable");
			}

			public ExpandableSubNet addGroup(String label, ExpandableSubNet parent) {
				throw new UnsupportedOperationException("Immutable");
			}

			public ExpandableSubNet addGroup(String label) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Arc addArc(Transition t, Place p, int weight, ExpandableSubNet parent) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Arc addArc(Place p, Transition t, int weight, ExpandableSubNet parent) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Arc addArc(Transition t, Place p, ExpandableSubNet parent) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Arc addArc(Place p, Transition t, ExpandableSubNet parent) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Arc addArc(Transition t, Place p, int weight) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Arc addArc(Place p, Transition t, int weight) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Arc addArc(Transition t, Place p) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Arc addArc(Place p, Transition t) {
				throw new UnsupportedOperationException("Immutable");
			}

			public void setGuardFromString(Transition transition, String guardAsString)
					throws ParseException, NonExistingVariableException {
				throw new UnsupportedOperationException("Immutable");
			}

			public void setGuard(Transition transition, GuardExpression guard) throws NonExistingVariableException {
				throw new UnsupportedOperationException("Immutable");
			}

			public void removeWriteOperation(Transition transition, DataElement variable) {
				throw new UnsupportedOperationException("Immutable");
			}

			public void removeReadOperation(Transition transition, DataElement variable) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Collection<DataElement> getVariables() {
				return ImmutableSet.of();
			}

			public DataElement getVariable(String name) {
				return null;
			}

			public void assignWriteOperation(Transition transition, DataElement variable) {
				throw new UnsupportedOperationException("Immutable");
			}

			public void assignReadOperation(Transition transition, DataElement variable) {
				throw new UnsupportedOperationException("Immutable");
			}

			@SuppressWarnings("rawtypes")
			public DataElement addVariable(String varName, Class type, Comparable minValue, Comparable maxValue) {
				throw new UnsupportedOperationException("Immutable");
			}

			public void setInitialMarking(Marking marking) {
				throw new UnsupportedOperationException("Immutable");
			}

			public void setFinalMarkings(Marking[] markings) {
				throw new UnsupportedOperationException("Immutable");
			}

			public Marking getInitialMarking() {
				return initialMarking;
			}

			public Marking[] getFinalMarkings() {
				return new Marking[] { finalMarking };
			}
		}

		private Factory(){
			//static
		}

		/**
		 * @param label
		 * @return a new {@link DataPetriNetsWithMarkings}
		 */
		public static DataPetriNetsWithMarkings create(String label) {
			return new PetriNetWithData(label);
		}

		/**
		 * @param net
		 * @return a copy of the {@link PetrinetGraph} as
		 *         {@link DataPetriNetsWithMarkings}
		 */
		public static DataPetriNetsWithMarkings fromPetrinet(PetrinetGraph net) {
			PetriNetWithDataFactory factory = new PetriNetWithDataFactory(net, net.getLabel());
			return factory.getRetValue();
		}

		/**
		 * Copies the {@link DataPetriNet} to a {@link Petrinet}, i.e., without
		 * guard, variables. This might change the behavior of the net, guards
		 * and variables are <B>not</B> converted
		 * 
		 * @param net
		 * @return
		 */
		public static Petrinet toPetrinet(DataPetriNet net) {
			PetrinetWithMarkings netWithMarkings = toPetrinetWithMarkings(net);
			return netWithMarkings.getNet();
		}

		/**
		 * Copies the {@link DataPetriNet} to a {@link Petrinet}, i.e., without
		 * guard, variables. This might change the behavior of the net, guards
		 * and variables are <B>not</B> converted
		 * <p>
		 * If the {@link DataPetriNet} is, in fact, a
		 * {@link DataPetriNetsWithMarkings}, then the initial and final marking
		 * is converted as well. Otherwise the returned markings are NULL.
		 * 
		 * @param net
		 * @return a copy as {@link Petrinet}
		 */
		public static PetrinetWithMarkings toPetrinetWithMarkings(DataPetriNet net) {
			final PetrinetImpl newNet = new PetrinetImpl(net.getLabel());
			Map<Place, Place> old2NewPlace = new HashMap<>();
			Map<Transition, Transition> old2NewTrans = new HashMap<>();
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : net.getEdges()) {
				if (edge.getSource() instanceof Transition && edge.getTarget() instanceof Place) {
					Place newPlace = addPlace((Place) edge.getTarget(), newNet, old2NewPlace);
					Transition newTrans = addTransition((Transition) edge.getSource(), newNet, old2NewTrans);
					newNet.addArc(newTrans, newPlace);
				} else if (edge.getSource() instanceof Place && edge.getTarget() instanceof Transition) {
					Place newPlace = addPlace((Place) edge.getSource(), newNet, old2NewPlace);
					Transition newTrans = addTransition((Transition) edge.getTarget(), newNet, old2NewTrans);
					newNet.addArc(newPlace, newTrans);
				} // else ignore as it is not part of the petrinet
			}
			if (net instanceof DataPetriNetsWithMarkings) {
				DataPetriNetsWithMarkings dpnWithMarking = (DataPetriNetsWithMarkings) net;
				final Marking initialMarking = dpnWithMarking.getInitialMarking() != null ? convert(dpnWithMarking.getInitialMarking(), old2NewPlace) : null;	
				final Marking[] finalMarkings = dpnWithMarking.getFinalMarkings() != null ? convertFinalMarking(dpnWithMarking.getFinalMarkings(), old2NewPlace) : null;
				return new PetrinetWithMarkings() {

					public Petrinet getNet() {
						return newNet;
					}

					public Marking getInitialMarking() {
						return initialMarking;
					}

					public Marking[] getFinalMarkings() {
						return finalMarkings;
					}
				};
			} else {
				return new PetrinetWithMarkings() {

					public Petrinet getNet() {
						return newNet;
					}

					public Marking getInitialMarking() {
						return null;
					}

					public Marking[] getFinalMarkings() {
						return null;
					}
				};
			}
		}

		private static Marking[] convertFinalMarking(Marking[] finalMarkings, Map<Place, Place> old2NewPlace) {
			if (finalMarkings != null && finalMarkings.length > 0) {
				Marking finalMarking = convert(finalMarkings[0], old2NewPlace);
				return new Marking[] { finalMarking };
			} else {
				return null;
			}
		}

		private static Marking convert(Marking marking, Map<Place, Place> old2NewPlace) {
			Marking newMarking = new Marking();
			for (Place p : marking) {
				Place place = old2NewPlace.get(p);
				if (place != null) {
					newMarking.add(place);
				} else {
					throw new IllegalArgumentException(
							"Place " + p + " is not part of the original net. Known mapping " + old2NewPlace);
				}
			}
			return newMarking;
		}

		private static Place addPlace(Place place, PetrinetImpl newNet, Map<Place, Place> old2NewPlace) {
			Place newPlace = old2NewPlace.get(place);
			if (newPlace == null) {
				newPlace = newNet.addPlace(place.getLabel());
			}
			old2NewPlace.put(place, newPlace);
			return newPlace;
		}

		private static Transition addTransition(Transition transition, PetrinetImpl newNet,
				Map<Transition, Transition> old2NewTrans) {
			Transition newTransition = old2NewTrans.get(transition);
			if (newTransition == null) {
				newTransition = newNet.addTransition(transition.getLabel());
				newTransition.setInvisible(transition.isInvisible());
			}
			old2NewTrans.put(transition, newTransition);
			return newTransition;
		}

		/**
		 * @param net
		 * @return an immutable view of the {@link PetrinetGraph} as
		 *         {@link DataPetriNet}
		 */
		public static DataPetriNet viewAsDataPetriNet(PetrinetGraph net) {
			return new ImmutablePetriNetAsDPN(net);
		}
		
		
		/**
		 * @param net
		 * @return an immutable view of the {@link PetrinetGraph} as
		 *         {@link DataPetriNet}
		 */
		public static DataPetriNet viewAsDataPetriNet(PetrinetGraph net, PluginContext context) {
			Marking initialMarking = null;
			Marking finalMarking = null;
			try {
				InitialMarkingConnection initConn = ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), InitialMarkingConnection.class, net);
				initialMarking = (Marking) initConn.getObjectWithRole(initConn.MARKING);
			} catch (ConnectionCannotBeObtained e) {
			}
			try {
				FinalMarkingConnection finalConn = ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), FinalMarkingConnection.class, net);				
				finalMarking = (Marking) finalConn.getObjectWithRole(finalConn.MARKING);
			} catch (ConnectionCannotBeObtained e) {
			}
			return new ImmutablePetriNetAsDPN(net, initialMarking, finalMarking);
		}
		
		
		

	}

	@SuppressWarnings("rawtypes")
	//TODO introduce generics
	DataElement addVariable(String varName, Class type, Comparable minValue, Comparable maxValue);

	DataElement getVariable(String name);

	Collection<DataElement> getVariables();

	void assignWriteOperation(Transition transition, DataElement variable);

	void removeWriteOperation(Transition transition, DataElement variable);

	void assignReadOperation(Transition transition, DataElement variable);

	void removeReadOperation(Transition transition, DataElement variable);

	void setGuardFromString(Transition transition, String guardAsString)
			throws ParseException, NonExistingVariableException;

	void setGuard(Transition transition, GuardExpression guard) throws NonExistingVariableException;

}
