package org.processmining.datapetrinets.conversion;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNet.PetrinetWithMarkings;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithDataFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.petrinets.utils.PetriNetUtils;
import org.processmining.plugins.utils.ConnectionManagerHelper;

public class DataPetriNetConversionsPlugin {

	@Plugin(name = "Convert Petri net to Data Petri net", level = PluginLevel.Regular, parameterLabels = {
			"Petri Net" }, returnLabels = { "Data Petri Net" }, returnTypes = {
					DataPetriNet.class }, userAccessible = true, keywords = { "DPN",
							"conversion" }, help = "Convert a Petrinet to a DataPetriNet, for example to use it in plug-ins that only accept a DataPetriNet.")

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public static DataPetriNet convertFromPetriNet(UIPluginContext context, PetrinetGraph net) {
		PetriNetWithDataFactory factory = new PetriNetWithDataFactory(net, net.getLabel());
		DataPetriNetsWithMarkings dpn = factory.getRetValue();
		try {
			InitialMarkingConnection initConn = ConnectionManagerHelper
					.safeGetFirstConnection(context.getConnectionManager(), InitialMarkingConnection.class, net);
			dpn.setInitialMarking(factory.convertMarking((Marking) initConn.getObjectWithRole(initConn.MARKING)));
		} catch (ConnectionCannotBeObtained e) {
		}
		try {
			FinalMarkingConnection finalConn = ConnectionManagerHelper
					.safeGetFirstConnection(context.getConnectionManager(), FinalMarkingConnection.class, net);
			dpn.setFinalMarkings(
					new Marking[] { factory.convertMarking((Marking) finalConn.getObjectWithRole(finalConn.MARKING)) });
		} catch (ConnectionCannotBeObtained e) {
		}
		return dpn;
	}

	@Plugin(name = "Convert Data Petri net to Petri net", level = PluginLevel.Regular, parameterLabels = {
			"Data Petri Net" }, returnLabels = { "Petri net" }, returnTypes = {
					Petrinet.class }, userAccessible = true, keywords = { "DPN",
							"conversion" }, help = "Convert a Petrinet to a DataPetriNet, for example to use it in plug-ins that only accept a DataPetriNet.")

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public static Petrinet convertToPetriNet(UIPluginContext context, DataPetriNet dpn) {
		PetrinetWithMarkings pnWithMarkings = DataPetriNet.Factory.toPetrinetWithMarkings(dpn);

		if (pnWithMarkings.getInitialMarking() != null) {
			PetriNetUtils.addInitialMarking(context, pnWithMarkings.getNet(), pnWithMarkings.getInitialMarking());
		}
		if (pnWithMarkings.getFinalMarkings() != null && pnWithMarkings.getFinalMarkings().length > 0
				&& pnWithMarkings.getFinalMarkings()[0] != null) {
			PetriNetUtils.addFinalMarking(context, pnWithMarkings.getNet(), pnWithMarkings.getFinalMarkings()[0]);
		}
		return pnWithMarkings.getNet();
	}

}
