package org.processmining.datapetrinets.io;

import java.io.IOException;
import java.io.InputStream;

import org.processmining.log.utils.XUtils;
import org.processmining.plugins.pnml.base.FullPnmlElementFactory;
import org.processmining.plugins.pnml.base.Pnml;
import org.processmining.plugins.pnml.base.PnmlElementFactory;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

class PnmlImport {

	public static Pnml importPnml(InputStream input) throws XmlPullParserException, IOException {
		/*
		 * Get an XML pull parser.
		 */
		XmlPullParserFactory xFactory = XmlPullParserFactory.newInstance();
		xFactory.setNamespaceAware(true);
		XmlPullParser xpp = xFactory.newPullParser();
		/*
		 * Initialize the parser on the provided input.
		 */
		xpp.setInput(input, null);
		/*
		 * Get the first event type.
		 */
		int eventType = xpp.getEventType();
		/*
		 * Create a fresh PNML object.
		 */
		PnmlElementFactory factory = new FullPnmlElementFactory();
		Pnml pnml = new Pnml();

		synchronized (factory) {
			pnml.setFactory(factory);
			/*
			 * Skip whatever we find until we've found a start tag.
			 */
			while (eventType != XmlPullParser.START_TAG) {
				eventType = xpp.next();
			}
			/*
			 * Check whether start tag corresponds to PNML start tag.
			 */
			if (xpp.getName().equals(Pnml.TAG)) {
				/*
				 * Yes it does. Import the PNML element.
				 */
				pnml.importElement(xpp, pnml);
			} else {
				/*
				 * No it does not. Return null to signal failure.
				 */
				pnml.log(Pnml.TAG, xpp.getLineNumber(), "Expected pnml");
			}
			if (pnml.hasErrors()) {
				throw new IOException(XUtils.stringifyLog(pnml.getLog()));
			}
		}
		return pnml;
	}

}
