package org.processmining.datapetrinets.ui.create;

import java.util.Collection;

import javax.swing.DefaultCellEditor;
import javax.swing.JTable;
import javax.swing.JTextField;

import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.syntax.ParseException;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;

public class AddGuardCellEditor extends DefaultCellEditor {

	private static final long serialVersionUID = 7082238577481966564L;

	private AddOperationsAndGuardsPanel frame;
	private JTable table;
	private Collection<DataElement> allDataElements;

	public AddGuardCellEditor(JTable table, Collection<DataElement> allDataElements, AddOperationsAndGuardsPanel frame) {
		super(new JTextField());
		this.table = table;
		this.frame = frame;
		this.allDataElements = allDataElements;
	}

	@Override
	public Object getCellEditorValue() {
		try {
			if (!((String) super.getCellEditorValue()).trim().isEmpty()) {
				GuardExpression expression = GuardExpression.Factory.newInstance((String) super.getCellEditorValue());
				for (String variableInExpression : expression.getNormalVariables()) {
					if (!checkNormalVariable(variableInExpression)) {
						frame.setMessage("Variable " + variableInExpression + " is not written by any other transition");
						throw new IllegalArgumentException("Variable " + variableInExpression
								+ " is not written by any other transition");
					}
				}
				for (String variableInExpression : expression.getPrimeVariables()) {
					if (!checkPrimeVariable(variableInExpression)) {
						frame.setMessage("Variable " + variableInExpression + " is not written by this transition");
						throw new IllegalArgumentException("Variable " + variableInExpression
								+ " is not written by this transition");
					}
				}
				frame.setMessage("Correctly parsed as " + expression.toCanonicalString() + ".");
			} else
				frame.setMessage("Correctly parsed as no guard");
			return super.getCellEditorValue();
		} catch (ParseException e) {
			frame.setMessage("Expression " + super.getCellEditorValue() + " is not parseable: " + e.getMessage());
			throw new IllegalArgumentException(e);
		}
	}

	private boolean checkNormalVariable(String variable) {
		if (allDataElements != null) {
			for (DataElement elem : allDataElements)
				if (elem.getVarName().equals(variable))
					return true;
		}
		return false;
	}

	private boolean checkPrimeVariable(String variable) {
		int row = table.getSelectedRow();
		@SuppressWarnings("unchecked")
		Collection<DataElement> dataElements = (Collection<DataElement>) table.getModel().getValueAt(row, 2);
		if (dataElements != null) {
			for (DataElement elem : dataElements) {
				if (elem.getVarName().equals(variable))
					return true;
			}
		}
		return false;
	}

}
