package org.processmining.datapetrinets.ui.create;

import java.awt.BorderLayout;
import java.text.ParseException;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;

import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.table.DefaultTableModel;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.widgets.wizard.AbstractDialog;
import org.processmining.widgets.wizard.Dialog;

class AddOperationsAndGuardsPanel extends AbstractDialog<PetriNetWithData> {
	private static final long serialVersionUID = 2096141300686063634L;

	private Object columnNames[] = { "Transition", "Variables Read", "Variable Written", "Guard" };
	private DefaultTableModel dtm;
	private JTable table;
	private Object data[][];
	private JTextArea errorMessageLbl;

	public void setMessage(String text) {
		errorMessageLbl.setText(text);
		errorMessageLbl.setCaretPosition(0);
	}

	public AddOperationsAndGuardsPanel(Map<Transition, Transition> map, Collection<DataElement> dataElements, 
			UIPluginContext context, 
			PetriNetWithData petrinet, 
			Dialog<PetriNetWithData> previous) {
		super(context,"Add Read/Write Operations and Guards",petrinet,previous);
		
		data = new Object[map.entrySet().size()][4];
		int index = 0;
		Set<String> varAsStringSet=new HashSet<String>();
		for(DataElement elem : dataElements)
			varAsStringSet.add(elem.getVarName());
		TreeSet<Entry<Transition, Transition>> t = new TreeSet<Entry<Transition, Transition>>(
				new Comparator<Entry<Transition, Transition>>() {

					public int compare(Entry<Transition, Transition> o1, Entry<Transition, Transition> o2) {
						if (o1.getValue().getLabel().equals(o2.getValue().getLabel()))
							return o1.getValue().getId().compareTo(o2.getValue().getId());
						else
							return o1.getValue().getLabel().compareTo(o2.getValue().getLabel());
					}
				});
		t.addAll(map.entrySet());
		for (Entry<Transition, Transition> entry : t) {
			Transition newTrans = entry.getValue();
			data[index][0] = newTrans;
			if (entry.getKey() instanceof PNWDTransition) {
				PNWDTransition oldTrans = (PNWDTransition) entry.getKey();
				if (oldTrans != null) {
					data[index][1] = retrieveFromCollection(oldTrans.getReadOperations(), dataElements);
					data[index][2] = retrieveFromCollection(oldTrans.getWriteOperations(), dataElements);
					if (oldTrans.getGuardExpression()!=null && 
							varAsStringSet.containsAll(oldTrans.getGuardExpression().getNormalVariables()) &&
							varAsStringSet.containsAll(oldTrans.getGuardExpression().getPrimeVariables()))
					{
						data[index][3] = oldTrans.getGuardAsString();
					}
					else
						data[index][3] = null;
				}
			}
			index++;
		}

		dtm = new DefaultTableModel(data, columnNames) {

			private static final long serialVersionUID = -564606983209878874L;

			public boolean isCellEditable(int row, int column) {
				return (column > 0);
			}
		};
		table = new JTable(dtm);
		table.getColumnModel().getColumn(1).setCellEditor(new AddVariablesCellEditor(dataElements));
		table.getColumnModel().getColumn(2).setCellEditor(new AddVariablesCellEditor(dataElements));
		table.getColumnModel().getColumn(3).setCellEditor(new AddGuardCellEditor(table, dataElements, this));
		this.setLayout(new BorderLayout());
		this.add(new JScrollPane(table), BorderLayout.CENTER);
		errorMessageLbl = new JTextArea();
		errorMessageLbl.setEditable(false);
		errorMessageLbl.setRows(6);
		this.add(new JScrollPane(errorMessageLbl), BorderLayout.SOUTH);
	}

	private Collection<DataElement> retrieveFromCollection(Set<DataElement> readOperations,
			Collection<DataElement> dataElements) {
		Collection<DataElement> retValue = new HashSet<DataElement>();
		for (DataElement elemToSearch : readOperations)
			for (DataElement elemToAdd : dataElements)
				if (elemToSearch.equals(elemToAdd)) {
					retValue.add(elemToAdd);
					break;
				}
		if (retValue.size() == 0)
			return null;
		else
			return retValue;
	}

	@SuppressWarnings("unchecked")
	public OperationsAndGuards[] getDataElement() {
		OperationsAndGuards[] retValue = new OperationsAndGuards[dtm.getRowCount()];
		for (int i = 0; i < dtm.getRowCount(); i++) {
			retValue[i] = new OperationsAndGuards((Transition) dtm.getValueAt(i, 0),
					(Collection<DataElement>) dtm.getValueAt(i, 1), (Collection<DataElement>) dtm.getValueAt(i, 2),
					(String) dtm.getValueAt(i, 3));
		}
		return retValue;
	}
	
	public boolean hasNextDialog() {
		return false;
	}

	public void updateParameters() {

		
	}

	public JComponent visualize() {
		return this;
	}

	protected boolean canProceedToNext() {
		if (table.isEditing())
			return false;
		try{
		for (OperationsAndGuards elem : getDataElement()) {
			if (elem.getGuard() != null && !elem.getGuard().isEmpty())
				getParameters().setGuard(elem.getTransition(), elem.getGuard());
			if (elem.getReadVariables() != null)
				for (DataElement dataElem : elem.getReadVariables())
					getParameters().assignReadOperation(elem.getTransition(), dataElem);
			if (elem.getWrittenVariables() != null)
				for (DataElement dataElem : elem.getWrittenVariables())
					getParameters().assignWriteOperation(elem.getTransition(), dataElem);
		}	
		}
		catch(ParseException ex)
		{
			return false;
		}
		return true;
	}

	protected Dialog<PetriNetWithData> determineNextDialog() {
		return null;
	}

	protected String getUserInputProblems() {
		return "Please complete the editing before proceeding (e.g., by pressing ENTER)";
	}

	public void updateParametersOnGetNext() {
		
	}

	public void updateParametersOnGetPrevious() {
		
	}
}
