package org.processmining.datapetrinets.ui.create;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.lang.reflect.Constructor;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.table.DefaultTableModel;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithDataFactory;
import org.processmining.widgets.wizard.AbstractDialog;
import org.processmining.widgets.wizard.Dialog;

import com.fluxicon.slickerbox.components.SlickerButton;

public class AddVariablesPanel extends AbstractDialog<PetriNetWithData> 
{
	
	private static final class CellValueVerification extends DefaultCellEditor {
		
		private static final long serialVersionUID = 5735048611210887511L;

		private final JTable table;
		private final JTextField field;

		public CellValueVerification(JTable table) {
			this(table, new JTextField());
		}

		private CellValueVerification(JTable table, JTextField aField) {
			super(aField);
			this.field = aField;
			this.table = table;
		}

		@Override
		public Object getCellEditorValue() {
			String value = field.getText();
			if (value==null || value.trim().equals(""))
				return value;
			Object convValue = null;
			Class<?> type = (Class<?>) table.getModel().getValueAt(table.getEditingRow(), 1);
			if (!value.isEmpty()) {
				try {
					convValue = type.getConstructor(String.class).newInstance(value);
				} catch (Exception err) {
					JOptionPane.showMessageDialog(null, "The value is not admissible for the data type", "Incomplete Parameter Selection",
							JOptionPane.WARNING_MESSAGE);
					throw new RuntimeException(err);
				}
			}
			if (type!=Date.class)
				return 	convValue.toString();
			else
				return value;
		}

	}
	
	private static final long serialVersionUID = 1L;
	private Object columnNames[] = { "Variable Name", "Type", "Minimum Value", "Maximum Value" };
	private DefaultTableModel dtm = new DefaultTableModel(columnNames, 0);
	private SlickerButton addButton = new SlickerButton("Add Variable");
	private SlickerButton removeButton = new SlickerButton("Remove Variable");
	private JTable table = new JTable(dtm);
	private PetriNetWithDataFactory factory;
	private String userInput;

	public AddVariablesPanel(PetriNetWithData retValue, UIPluginContext context, PetriNetWithDataFactory factory) {
		super(context,"Add Process Variables",retValue,null);
		init();
		this.factory=factory;
	}

	private void init() {
		JComboBox<Object> combo = new JComboBox<Object>();
		combo.addItem(String.class);
		combo.addItem(Long.class);
		combo.addItem(Double.class);
		combo.addItem(Date.class);
		combo.addItem(Boolean.class);
		table.getColumnModel().getColumn(3).setCellEditor(new CellValueVerification(table));
		table.getColumnModel().getColumn(2).setCellEditor(new CellValueVerification(table));
		table.getColumnModel().getColumn(1).setCellEditor(new DefaultCellEditor(combo));

		this.setLayout(new BorderLayout());
		this.add(new JScrollPane(table), BorderLayout.CENTER);
		JPanel south = new JPanel(new FlowLayout(FlowLayout.CENTER));
		south.add(addButton);
		south.add(removeButton);
		this.add(south, BorderLayout.SOUTH);

		addButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				dtm.addRow((Object[]) null);
			}
		});

		removeButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				int indexes[] = table.getSelectedRows();
				int confirm=JOptionPane.NO_OPTION;
				if (indexes.length>0)
				{
					confirm=JOptionPane.showConfirmDialog(null, "By deleting one or more variable(s), also the guards that contain"
							+ " any of that/those variable(s) will also be removed. Do you confirm?", "Remove variables", JOptionPane.YES_NO_OPTION);
				}
				if (confirm==JOptionPane.YES_OPTION)
				{
					for (int i = indexes.length - 1; i >= 0; i--)
						dtm.removeRow(indexes[i]);
				}
			}
		});
	}

	public AddVariablesPanel(DataPetriNet inputNet, PetriNetWithData retValue, UIPluginContext context, PetriNetWithDataFactory factory) {
		super(context,"Add Process Variables",retValue,null);
		this.factory=factory;
		DataElement[] variables = inputNet.getVariables().toArray(new DataElement[0]);
		Arrays.sort(variables, new Comparator<DataElement>() {

			public int compare(DataElement o1, DataElement o2) {
				return o1.getVarName().compareTo(o2.getVarName());
			}
		});
		for (DataElement elem : variables) {
			Object row[] = new Object[4];
			row[0] = elem.getVarName();
			row[1] = elem.getType();
			Comparable value = elem.getMinValue();
			if (value != null)
				row[2] = value.toString();
			else
				row[2] = "";
			value = elem.getMaxValue();
			if (value != null)
				row[3] = value.toString();
			else
				row[3] = "";
			dtm.addRow(row);
		}
		init();
	}

	@SuppressWarnings("unchecked")
	public Set<DataElement> getDataElements() {
		Set<DataElement> retValue;
		//if (retValue != null)
		//	return retValue;
		int index = dtm.getRowCount() - 1;
		retValue = new HashSet<DataElement>();
		for (; index >= 0; index--) {
			DataElement element;
			String varName = (String) dtm.getValueAt(index, 0);
			Class varType = (Class) dtm.getValueAt(index, 1);
			String minValueString = (String) dtm.getValueAt(index, 2);
			if (minValueString != null && minValueString.equals(""))
				minValueString = null;
			String maxValueString = (String) dtm.getValueAt(index, 3);
			if (maxValueString != null && maxValueString.equals(""))
				maxValueString = null;
			Constructor constr = null;
			Comparable<Object> minValue = null;
			Comparable<Object> maxValue = null;
			try {
				constr = varType.getConstructor(String.class);
				if (minValueString != null)
					minValue = (Comparable<Object>) constr.newInstance(minValueString);
				if (maxValueString != null)
					maxValue = (Comparable<Object>) constr.newInstance(maxValueString);
				element = new DataElement(varName, varType, minValue, maxValue, null);
				retValue.add(element);
			} catch (Exception e) {
				e.printStackTrace();
				if (varName==null)
					throw(new IllegalArgumentException("One of the variables has no name"));
				else if (varType==null) 
					throw(new IllegalArgumentException("Variable "+varName+" has no type"));
				else if (maxValue!=null && minValue!=null && maxValue.compareTo(minValue)<0)
					throw(new IllegalArgumentException("For variable "+varName+", the maximum value is smaller than the minimum value"));									
				else 
					throw(new IllegalArgumentException("Impossible to parse variable "+varName+". For example: is the minimum or maximum value of an "
							+ "incorrect type?"));
			}

		}
		return retValue;

	}

	public boolean hasNextDialog() {
		return true;
	}

	public void updateParameters() {	
		getParameters().removeAllVariables();
		for (DataElement elem : getDataElements())
			getParameters().addVariable(elem.getVarName(), elem.getType(), elem.getMinValue(), elem.getMaxValue());

	}

	public JComponent visualize() {
		return this;
	}

	protected boolean canProceedToNext() { 
		if (table.isEditing())
		{
			userInput="Complete the editing of the data variable before going to the next";
			return false;
		}
		try {
			getDataElements();
			return true;
		} catch (IllegalArgumentException e) {
			userInput=e.getMessage();
			return false;
		}
	}

	protected Dialog<PetriNetWithData> determineNextDialog() {
		return new AddOperationsAndGuardsPanel(factory.getTransMapping(),getParameters().getVariables(), 
				getUIPluginContext(), getParameters(), this);
	}

	protected String getUserInputProblems() {
		return userInput;
	}

	public void updateParametersOnGetNext() {
		updateParameters();
	}

	public void updateParametersOnGetPrevious() {
		updateParameters();
	}

}