package org.processmining.datapetrinets.validation;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.cpntools.accesscpn.engine.highlevel.instance.Binding;
import org.cpntools.accesscpn.engine.highlevel.instance.State;
import org.cpntools.accesscpn.model.Transition;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.framework.util.Pair;
import org.processmining.models.semantics.petrinet.Marking;



public class CPNAnalysisResults {
	Set<Transition> deadTransitions=new HashSet<Transition>();
	List<Pair<State,Binding[]>> deadlocks=new LinkedList<Pair<State,Binding[]>>();
	HashSet<HashMap<String,String>> addedStates=new HashSet<HashMap<String,String>>();
	
	private final DataPetriNet dpn;
	private final Marking initMarking;
	private final Marking finalMarking;

	public CPNAnalysisResults(DataPetriNet dpn, org.processmining.models.semantics.petrinet.Marking initMarking,
			org.processmining.models.semantics.petrinet.Marking finalMarking) {
		this.dpn=dpn;
		this.initMarking=initMarking;
		this.finalMarking=finalMarking;
	}

	public void addDeadTransition(Transition t) {
		deadTransitions.add(t);
	}

	public void addDeadlock(State deadlock, Binding[] path) 
	{
		
		HashMap<String,String> elem=new HashMap<String,String>();
		for(org.cpntools.accesscpn.engine.highlevel.instance.Marking m : deadlock.getAllMarkings())
		{
			String placeName=m.getPlaceInstance().getNode().getName().asString();
			if (!placeName.startsWith("Variable") && !placeName.startsWith("Potential"))
			{
				elem.put(placeName, m.getMarking());
			}
		}
		if (!addedStates.contains(elem))
		{
			deadlocks.add(new Pair<State, Binding[]>(deadlock, path));
			addedStates.add(elem);
		}
	}
	
	public Set<Transition> getDeadTransitions()
	{
		return deadTransitions;
	}

	public List<Pair<State, Binding[]>> getDeadlocks() {
		return deadlocks;
	}

	public DataPetriNet getDpn() {
		return dpn;
	}

	public Marking getInitMarking() {
		return initMarking;
	}

	public Marking getFinalMarking() {
		return finalMarking;
	}

}
