package org.processmining.datapetrinets.validation;

import java.awt.BorderLayout;

import javax.swing.DefaultListModel;
import javax.swing.JComponent;
import javax.swing.JSplitPane;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.cpntools.accesscpn.engine.highlevel.instance.Binding;
import org.cpntools.accesscpn.engine.highlevel.instance.ValueAssignment;
import org.cpntools.accesscpn.model.Transition;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.util.ui.widgets.ProMList;
import org.processmining.framework.util.ui.widgets.ProMTable;

import com.fluxicon.slickerbox.components.RoundedPanel;
import com.fluxicon.slickerbox.factory.SlickerFactory;

public class CPNAnalysisResultsVisualizer {
	
	@Plugin(
			name="Prediction", 
			level=PluginLevel.NightlyBuild,
			parameterLabels={"AnalysisResults"}, 
			returnLabels={"JComponent"}, 
			returnTypes={JComponent.class}, 
			userAccessible=true
			)
	@Visualizer

	@UITopiaVariant(
			affiliation = "TU/e", 
	        author = "Massimiliano", 
	        email = "m.d.leoni@tue.nl")
	public JComponent visualizePrediction(PluginContext context, CPNAnalysisResults analysisResults) throws Exception
	{
		
		final DefaultListModel<String> listModelDeadTransitions = new DefaultListModel<String>();
			
		for(Transition t : analysisResults.getDeadTransitions())
			listModelDeadTransitions.addElement(t.getName().asString());

		ProMList<String> deadTransitionsList=new ProMList<String>("Dead Transitions",listModelDeadTransitions);
		
		RoundedPanel deadLockPanel=new RoundedPanel();
		
		deadLockPanel.setLayout(new BorderLayout());
		
		deadLockPanel.add(SlickerFactory.instance().createLabel("Deadlocks:"),BorderLayout.NORTH);
		
		final DefaultListModel<String> listModelTransitionFirings = new DefaultListModel<String>();
		
		final ProMList<String> transitionFiring=new ProMList<String>("Example of Execution Leading to the Deadlock",listModelTransitionFirings);
		
		final DeadLockTableModel tm = new DeadLockTableModel(analysisResults.getDeadlocks());
		
		final ProMTable deadlockTable=new ProMTable(tm);
		deadLockPanel.add(deadlockTable,BorderLayout.CENTER);
		deadlockTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		deadlockTable.getSelectionModel().addListSelectionListener(new ListSelectionListener()
				{

					public void valueChanged(ListSelectionEvent arg0) {
						int row=deadlockTable.getSelectedRow();
					       if (row>=0)
					        {
					        	listModelTransitionFirings.clear();
					        	for (Binding b : tm.getBinding(row))
					        	{
					        		String name=b.getTransitionInstance().getNode().getName().asString();
					        		name=name.substring(0, name.lastIndexOf('_'));
					        		StringBuffer sb=new StringBuffer();
					        		sb.append(name);
					        		sb.append(" { ");
					        		boolean theFirst=true;
					        		for(ValueAssignment ass : b.getAllAssignments())
					        		{
					        			if (ass.getName().contains("Prime"))
					        			{
					        				if(!theFirst)
					        					sb.append(", ");
					        				sb.append(ass.getName().replace("Prime", ""));
					        				sb.append(" = ");
					        				sb.append(ass.getValue());
					        				theFirst=false;
					        			}
					        		}
					        		sb.append(" }");
					        		listModelTransitionFirings.addElement(sb.toString());
					        		
					        	}
					        }						
					}});
	
		JSplitPane split=new JSplitPane(JSplitPane.VERTICAL_SPLIT,deadTransitionsList,deadLockPanel);
		split.setResizeWeight(0.1);
		JSplitPane split2=new JSplitPane(JSplitPane.VERTICAL_SPLIT,split,transitionFiring);
		split2.setResizeWeight(0.7);
		
		return split2;
		
	}
}
