package org.processmining.datapetrinets.validation;

import java.awt.geom.Point2D;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.swing.JOptionPane;

import org.cpntools.accesscpn.model.ModelFactory;
import org.cpntools.accesscpn.model.Name;
import org.cpntools.accesscpn.model.Page;
import org.cpntools.accesscpn.model.PetriNet;
import org.cpntools.accesscpn.model.graphics.Coordinate;
import org.cpntools.accesscpn.model.graphics.GraphicsFactory;
import org.cpntools.accesscpn.model.graphics.NodeGraphics;
import org.cpntools.accesscpn.model.util.BuildCPNUtil;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.syntax.ExprAtLeast;
import org.processmining.datapetrinets.expression.syntax.ExprAtMost;
import org.processmining.datapetrinets.expression.syntax.ExprEqual;
import org.processmining.datapetrinets.expression.syntax.ExprGreaterThan;
import org.processmining.datapetrinets.expression.syntax.ExprLessThan;
import org.processmining.datapetrinets.expression.syntax.ExprLitDouble;
import org.processmining.datapetrinets.expression.syntax.ExprLitInteger;
import org.processmining.datapetrinets.expression.syntax.ExprLitString;
import org.processmining.datapetrinets.expression.syntax.ExprNotEqual;
import org.processmining.datapetrinets.expression.syntax.ExprVariable;
import org.processmining.datapetrinets.expression.syntax.ExpressionParserDefaultVisitor;
import org.processmining.datapetrinets.expression.syntax.ExpressionVisitorException;
import org.processmining.datapetrinets.expression.syntax.SimpleNode;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.petrinet.finalmarkingprovider.MarkingEditorPanel;
import org.processmining.plugins.utils.ProvidedObjectHelper;

class AtomExtractor extends ExpressionParserDefaultVisitor
{
	private final DataPetriNet dpn;

	public AtomExtractor(DataPetriNet dpn) {
		this.dpn=dpn;
	}

	public void addNumbers(SimpleNode node, Object data)
	{
		if (node.jjtGetNumChildren()==2 && node.jjtGetChild(0) instanceof ExprVariable)
		{
			String varName=(String) ((ExprVariable)node.jjtGetChild(0)).jjtGetValue();
			varName=varName.replace('\'', ' ').trim();
			@SuppressWarnings("unchecked")
			HashMap<String, Set<Object>> table = (HashMap<String, Set<Object>>)data;
			Set<Object> listOfValues = table.get(varName);	
			if (dpn.getVariable(varName).getType()==Long.class)
			{
				String s=null;
				if (node.jjtGetChild(1) instanceof ExprLitInteger)
				{
					s=(String) ((ExprLitInteger)node.jjtGetChild(1)).jjtGetValue();
				}
				else if (node.jjtGetChild(1) instanceof ExprLitDouble)
					s=(String) ((ExprLitDouble)node.jjtGetChild(1)).jjtGetValue();
				else return;
				double value=Double.parseDouble(s);
				listOfValues.add(value);

			}
			else if (dpn.getVariable(varName).getType()==Double.class)
			{
				String s=null;
				if (node.jjtGetChild(1) instanceof ExprLitInteger)
					s=(String) ((ExprLitInteger)node.jjtGetChild(1)).jjtGetValue();
				else if (node.jjtGetChild(1) instanceof ExprLitDouble)
					s=(String) ((ExprLitDouble)node.jjtGetChild(1)).jjtGetValue();
				else return;			
				double value=Double.parseDouble(s);
				listOfValues.add(value);
				/*SortedSet<Object> list = listOfValues.tailSet(value);
				if (list.isEmpty())
					listOfValues.add(value+1);
				else 
				{
					Iterator<Object> iter = list.iterator();
					double valueInList=(double) iter.next();
					if(valueInList != value)
						listOfValues.add((valueInList - value)/2D);
					else 
					{
						if (iter.hasNext())
						{
							valueInList=(double) iter.next();
							listOfValues.add((valueInList - value)/2D);
						}
						else
							listOfValues.add(value+1);
					}
				}*/
			}
			else if (node.jjtGetChild(1) instanceof ExprLitString)
			{
				String s=(String) ((ExprLitString)node.jjtGetChild(1)).jjtGetValue();
				s=s.substring(1, s.length()-1);
				listOfValues.add(s);

			}
		}
		
	}

	public Object visit(ExprEqual node, Object data) throws ExpressionVisitorException {
		addNumbers(node,data);
		return super.visit(node, data);
	}
	
	public Object visit(ExprNotEqual node, Object data) throws ExpressionVisitorException {
		addNumbers(node,data);
		return super.visit(node, data);
	}

	public Object visit(ExprLessThan node, Object data) throws ExpressionVisitorException {
		addNumbers(node,data);
		return super.visit(node, data);
	}

	public Object visit(ExprAtMost node, Object data) throws ExpressionVisitorException {
		addNumbers(node,data);
		return super.visit(node, data);
	}

	public Object visit(ExprGreaterThan node, Object data) throws ExpressionVisitorException {
		addNumbers(node,data);
		return super.visit(node, data);
	}

	public Object visit(ExprAtLeast node, Object data) throws ExpressionVisitorException {
		addNumbers(node,data);
		return super.visit(node, data);
	}

	/*public Object visit(ExprNegation node, Object data) throws ExpressionVisitorException {
		addNumbers(node,data);
		return super.visit(node, data);
	}

	public Object visit(ExprNot node, Object data) throws ExpressionVisitorException {
		addNumbers(node,data);
		return super.visit(node, data);
	}*/
}


public class ConvertToCPNTools {

	@Plugin(name = " Convert Data Petri Nets to CPN", level = PluginLevel.Regular, parameterLabels = {
	"Petri Net" }, returnLabels = { "Colour Petri Net" }, returnTypes = {
			PetriNet.class }, userAccessible = true, keywords = { "DPN",
	"conversion" }, help = "Convert a Petrinet to a Colour Petri Net for Soundness Analysis.")

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "M. de Leoni", email = "m.d.leoni@tue.nl")
	public static PetriNet convertFromPetriNet(UIPluginContext context, DataPetriNet dpn) throws Exception  {
		BuildCPNUtil cpnUtil = new BuildCPNUtil();

		/* Creating the Colour Petri Net and the main page*/
		PetriNet cpnNet = cpnUtil.createPetriNet();
		
		
		Page page = cpnUtil.addPage(cpnNet, "RootPage");

		/* Creating the mapping structures between DPN Places/Transitions with CPN Places/Transitions */
		HashMap<Transition, org.cpntools.accesscpn.model.Transition> transitionMap=
				new HashMap<Transition, org.cpntools.accesscpn.model.Transition>();
		HashMap<Place, org.cpntools.accesscpn.model.Place> placeMap=new HashMap<Place, org.cpntools.accesscpn.model.Place>();
		Map<String,Set<Object>> observedValues=new HashMap<String, Set<Object>>();
		Map<String,org.cpntools.accesscpn.model.Place> variablePlaces=new HashMap<String, org.cpntools.accesscpn.model.Place>();
		Map<String,org.cpntools.accesscpn.model.Place> variableValuePlaces=new HashMap<String, org.cpntools.accesscpn.model.Place>();
		Map<String,String> variableCPNType=new HashMap<String, String>();
		
		cpnUtil.declareColorSet(cpnNet, "MYUNIT", cpnUtil.cf.createCPNUnit());
		cpnUtil.declareColorSet(cpnNet, "INT", cpnUtil.cf.createCPNInt());
		cpnUtil.declareColorSet(cpnNet, "STRING", cpnUtil.cf.createCPNString());
		cpnUtil.declareColorSet(cpnNet, "REAL", cpnUtil.cf.createCPNReal());
		cpnUtil.declareColorSet(cpnNet, "BOOL", cpnUtil.cf.createCPNBool());
		
		GraphLayoutConnection layout=null;
		try {
			layout = context.getConnectionManager().getFirstConnection(GraphLayoutConnection.class, context, dpn);
		} catch (ConnectionCannotBeObtained e) {
			System.err.println(e.getMessage());
		}		
		
		
		//CREATE THE VARIABLE PLACES
		int pos=20;
				
		for(DataElement var : dpn.getVariables())
		{
			String cpnType,defaultValue;
			if (var.getType() == String.class)
			{
				cpnType="STRING";
				defaultValue="\"\"";
				variableCPNType.put(var.getVarName(),"STRING");
			}
			else if (var.getType() == Boolean.class)
			{
				cpnType="BOOL";
				defaultValue="false";
				variableCPNType.put(var.getVarName(),"BOOL");
			}			
			else if (var.getType() == Long.class || var.getType() == Integer.class)
			{
				cpnType="REAL";
				defaultValue="0.0";
				variableCPNType.put(var.getVarName(),"REAL");
			}
			else if (var.getType() == Double.class || var.getType() == Float.class)
			{
				cpnType="REAL";
				defaultValue="0.0";
				variableCPNType.put(var.getVarName(),"REAL");
			}
			else
				continue;
			org.cpntools.accesscpn.model.Place cpnPlace = cpnUtil.addPlace(page, "Variable "+var.getVarName().replace('_', '-'), 
					cpnType,defaultValue);
			cpnUtil.declareVariable(cpnNet, var.getVarName(), cpnType);
			cpnUtil.declareVariable(cpnNet, var.getVarName()+"Prime", cpnType);			
			observedValues.put(var.getVarName(), new HashSet<Object>());
			variablePlaces.put(var.getVarName(), cpnPlace);
			

			if (layout!=null)
			{
				NodeGraphics nodeGraph = GraphicsFactory.INSTANCE.createNodeGraphics();
				Coordinate coordCPN = GraphicsFactory.INSTANCE.createCoordinate();
				Point2D coordDPN = layout.getPosition(var);
				coordCPN.setX(coordDPN.getX()*2);
				coordCPN.setY(coordDPN.getY()*2);
				nodeGraph.setPosition(coordCPN);
				cpnPlace.setGraphics(nodeGraph);
			}
			
		}
		
		//POPULATE THE OBSERVED VALUES
		for(Transition transition : dpn.getTransitions())
		{
			PNWDTransition dpnTransition=(PNWDTransition) transition;
			GuardExpression expression = dpnTransition.getGuardExpression();
			if (expression!=null)
				expression.visit(new AtomExtractor(dpn), observedValues);
		}

		
		for(Entry<String, Set<Object>> entry : observedValues.entrySet())
		{
			if (variableCPNType.get(entry.getKey()).equals("STRING"))
			{
				Set<Object> setOfStrings = entry.getValue();
				if (setOfStrings.size()>0)
				{
					String elem=(String)setOfStrings.iterator().next()+((char)(new Random().nextInt(21)+'a'));
					while(setOfStrings.contains(elem))
					{
						elem+=((char)(new Random().nextInt(21)+'a'));
					}
					setOfStrings.add(elem);
				}
			}
			else if (dpn.getVariable(entry.getKey()).getType()==Long.class || dpn.getVariable(entry.getKey()).getType()==Integer.class)
			{
				Set<Object> setOfElems = entry.getValue();
				if (setOfElems.size()>0)
				{				
					Object[] asArray = entry.getValue().toArray(new Object[entry.getValue().size()]);
					Arrays.sort(asArray);

					double currElem=(double)asArray[0];
					double nextElem;
					currElem-=0.6;
					setOfElems.add(((double)Math.round(currElem)));
					for(int i=0;i<asArray.length-1;i++)
					{
						currElem=(double)asArray[i];
						nextElem=(double)asArray[i+1];
						setOfElems.add(((double)Math.round((currElem+nextElem)/2)));

					}
					currElem=(double)asArray[asArray.length-1];
					if (currElem!=Math.round(currElem))
					{
						System.out.println("Removed elem "+currElem);
						setOfElems.remove(currElem);
					}
					currElem+=0.6;
					setOfElems.add(((double)Math.round(currElem)));			
				}
			}
			else if (dpn.getVariable(entry.getKey()).getType()==Double.class || dpn.getVariable(entry.getKey()).getType()==Float.class)
			{
				Set<Object> setOfElems = entry.getValue();
				if (setOfElems.size()>0)
				{				
					Object[] asArray = entry.getValue().toArray(new Object[entry.getValue().size()]);
					Arrays.sort(asArray);

					double currElem=(double)asArray[0];
					double nextElem;
					currElem-=0.6;
					if (currElem>=0)
						setOfElems.add(currElem);
					for(int i=0;i<asArray.length-1;i++)
					{
						currElem=(double)asArray[i];
						nextElem=(double)asArray[i+1];
						setOfElems.add((currElem+nextElem)/2.0);

					}
					currElem=(double)asArray[asArray.length-1];
					if (currElem!=Math.round(currElem))
					{
						System.out.println("Removed elem "+currElem);
						setOfElems.remove(currElem);
					}
					currElem+=0.6;
					setOfElems.add(((double)Math.round(currElem)));			
				}
			}			
		}
		
		
		//CREATE THE VARIABLE-VALUE PLACES
		
		pos=20;
		
		for(Entry<String, Set<Object>> entry : observedValues.entrySet())
		{
			Iterator<Object> iter = entry.getValue().iterator();
			
			StringBuffer initMarking;
			if (iter.hasNext())
			{
				if (variableCPNType.get(entry.getKey()).equals("STRING"))
					initMarking=new StringBuffer("1`\""+iter.next()+"\"");
				else if (variableCPNType.get(entry.getKey()).equals("REAL"))
				{
					Double val = ((Number)iter.next()).doubleValue();
					String valString=val.toString();
					if (val<0)
					{
						valString="(~"+valString.replace('-', '~')+")";
					}
					initMarking=new StringBuffer("1`"+valString);
				}
				else
					initMarking=new StringBuffer("1`"+((Number)iter.next()).intValue());
			}
			else
			{
				if (variableCPNType.get(entry.getKey()).equals("STRING"))
					initMarking=new StringBuffer("1`\"\"");
				else if (variableCPNType.get(entry.getKey()).equals("BOOL"))
					initMarking=new StringBuffer("1`false++1`true");
				else if (variableCPNType.get(entry.getKey()).equals("REAL"))
					initMarking=new StringBuffer("1`0.0");
				else
					initMarking=new StringBuffer("1`0.0");					
			}
			while(iter.hasNext())
			{
				if (variableCPNType.get(entry.getKey()).equals("STRING"))
					initMarking.append("++1`\""+iter.next()+"\"");
				else if (variableCPNType.get(entry.getKey()).equals("REAL"))
				{
					Double val = ((Number)iter.next()).doubleValue();
					String valString=val.toString();
					if (val<0)
					{
						valString="("+valString.replace('-', '~')+")";
					}					
					initMarking.append("++1`"+valString);
				}
			}
			
			org.cpntools.accesscpn.model.Place cpnPlace = cpnUtil.addPlace(page, "Potential Values for "+entry.getKey().replace('_', '-'), 
					variableCPNType.get(entry.getKey()),initMarking.toString());
			variableValuePlaces.put(entry.getKey(), cpnPlace);
			
			NodeGraphics nodeGraph = GraphicsFactory.INSTANCE.createNodeGraphics();
			Coordinate coordCPN = GraphicsFactory.INSTANCE.createCoordinate();

			coordCPN.setX(pos);
			coordCPN.setY(0);
			pos+=50;
			
			nodeGraph.setPosition(coordCPN);
			cpnPlace.setGraphics(nodeGraph);			

		}
		

		SortedSet<Place> initMarking;
		try {
			initMarking = ((Marking)context.getConnectionManager()
					.getFirstConnection(InitialMarkingConnection.class, context, dpn)
					.getObjectWithRole(InitialMarkingConnection.MARKING)).baseSet();		
			if (initMarking.isEmpty()) {
				int yn = JOptionPane.showConfirmDialog(null, "The initial marking is empty. Is this intended?",
						"Initial Marking", JOptionPane.YES_NO_OPTION);
				if (yn == JOptionPane.NO_OPTION) {
					throw(new IllegalArgumentException());
				}
			}
		} catch (IllegalArgumentException | ConnectionCannotBeObtained e) {
			initMarking = createMarking(context, dpn, "Initial Marking");
			Marking toPublish=new Marking(initMarking);
			ProvidedObjectHelper.publish(context, "Initial Marking for " + dpn.getLabel(), toPublish, Marking.class,
					false);
			context.getConnectionManager().addConnection(new InitialMarkingConnection(dpn, toPublish));
		}		

		int counter=0;
		for(Place dpnPlace : dpn.getPlaces())
		{
			String label=dpnPlace.getLabel();
			if (label.trim().equals(""))
				label="p";
			org.cpntools.accesscpn.model.Place cpnPlace = cpnUtil.addPlace(page, label.replace('-', '_')+"_"+(++counter), 		
					"MYUNIT",initMarking.contains(dpnPlace)? "1`()" : "");
			if (layout!=null)
			{
				NodeGraphics nodeGraph = GraphicsFactory.INSTANCE.createNodeGraphics();
				Coordinate coordCPN = GraphicsFactory.INSTANCE.createCoordinate();
				Point2D coordDPN = layout.getPosition(dpnPlace);
				coordCPN.setX(coordDPN.getX()*2);
				coordCPN.setY(coordDPN.getY()*2);
				nodeGraph.setPosition(coordCPN);
				cpnPlace.setGraphics(nodeGraph);
			}			
			placeMap.put(dpnPlace, cpnPlace);
		}
		
		counter=0;
			for(Transition dpnTransition : dpn.getTransitions())
			{
				String label=dpnTransition.getLabel();
				if (label.trim().equals(""))
					label="t";
				String guard=null;
				
				if (((PNWDTransition)dpnTransition).getGuard()!=null)
					guard=((PNWDTransition)dpnTransition).getGuardAsString();
				HashSet<String> writtenVariables=new HashSet<String>();
				for(DataElement elem : ((PNWDTransition)dpnTransition).getWriteOperations())
					writtenVariables.add(elem.getVarName());
				if (guard!=null && guard.trim().length()>0)
				{
					guard=guard.trim().replace("'", "Prime").replace("&&", " andalso ").replace("==", "=").replace("||", " orelse ").replace("!=", "<>");
					if (writtenVariables.contains("totalPaymentAmount"))
					{
							guard+=" andalso (totalPaymentAmountPrime >= totalPaymentAmount)";
					}
					if (writtenVariables.contains("paymentAmount"))
					{
							guard+=" andalso (paymentAmountPrime >= paymentAmount)";
					}					
					if (writtenVariables.contains("amount"))
					{
							guard+=" andalso (amountPrime >= amount)";
					}	
					guard="("+guard+")";
				}

				if(guard!=null)
					guard=guard.replace(".0", ".x").replace("0)", "0.0)").replace(".x", ".0");
				
				
				org.cpntools.accesscpn.model.Transition cpnTransition = cpnUtil.
						addTransition(
								page,
								label.replace('-', '_')+"_"+(++counter),
								
								guard);
				
				GuardExpression expr = ((PNWDTransition)dpnTransition).getGuardExpression();
				Set<String> variablesInGuard = null;
				if (expr!=null)
					variablesInGuard = ((PNWDTransition)dpnTransition).getGuardExpression().getNormalVariables();
				
				for(Entry<String, org.cpntools.accesscpn.model.Place> entry :variablePlaces.entrySet())
				{
					if (writtenVariables.contains(entry.getKey()))
					{
						cpnUtil.addArc(page, cpnTransition,entry.getValue(),entry.getKey()+"Prime");
						cpnUtil.addArc(page, entry.getValue(),cpnTransition,entry.getKey());
					}	
					else if (variablesInGuard!=null && variablesInGuard.contains(entry.getKey()))
					{
						cpnUtil.addArc(page, entry.getValue(),cpnTransition,entry.getKey());
						cpnUtil.addArc(page, cpnTransition,entry.getValue(),entry.getKey());
					}
				} 
				
				for(Entry<String, org.cpntools.accesscpn.model.Place> entry : variableValuePlaces.entrySet())
				{
					if (writtenVariables.contains(entry.getKey()))
					{
						cpnUtil.addArc(page, entry.getValue(),cpnTransition,entry.getKey()+"Prime");
						cpnUtil.addArc(page, cpnTransition,entry.getValue(),entry.getKey()+"Prime");
					}

				}				

				
				transitionMap.put(dpnTransition, cpnTransition);
				if (layout!=null)
				{
					Coordinate coordCPN = GraphicsFactory.INSTANCE.createCoordinate();
					Point2D coordDPN = layout.getPosition(dpnTransition);
					coordCPN.setX(coordDPN.getX()*2);
					coordCPN.setY(coordDPN.getY()*2);
					NodeGraphics nodeGraph = GraphicsFactory.INSTANCE.createNodeGraphics();
					nodeGraph.setPosition(coordCPN);
					cpnTransition.setGraphics(nodeGraph);
				}
			}
			
			 
			for(Transition dpnTransition : dpn.getTransitions())
			{
				org.cpntools.accesscpn.model.Transition cpnTransition = transitionMap.get(dpnTransition);
				for(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> dpnEdge : dpn.getInEdges(dpnTransition))
				{
					if (dpnEdge.getSource() instanceof Place)
						cpnUtil.addArc(page,placeMap.get(dpnEdge.getSource()),cpnTransition,"()");
				}
				for(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> dpnEdge : dpn.getOutEdges(dpnTransition))
				{
					if (dpnEdge.getTarget() instanceof Place)
						cpnUtil.addArc(page,cpnTransition,placeMap.get(dpnEdge.getTarget()),"()");
				}			
			}
			
			
			/*File app=File.createTempFile("cpn", String.valueOf(System.currentTimeMillis()));
			FileOutputStream fos=new FileOutputStream(app);
			
			org.cpntools.accesscpn.model.exporter.DOMGenerator.export(cpnNet, fos);
			
			fos.flush();
			fos.close();*/
			
			//simulate(app);
			//cpnNet.setName(new Name("Colour Petri Net of "+dpn.getLabel());
			Name name = ModelFactory.INSTANCE.createName();
			name.setText("Colour Petri Net of "+dpn.getLabel());
			cpnNet.setName(name);
			
			return cpnNet;

		}

	public static SortedSet<Place> createMarking(PluginContext context, DataPetriNet dpn, String title) {
		MarkingEditorPanel editor = new MarkingEditorPanel(title);

		Marking aux = editor.getMarking((UIPluginContext) context, DataPetriNet.Factory.toPetrinetWithMarkings(dpn).getNet());
		return new TreeSet<Place>(aux);
	}


	}
