package org.processmining.datapetrinets.visualization.graphviz;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.utils.MarkingsHelper;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;

@Plugin(name = "Convert DPN-net to Graphviz DOT", returnLabels = { "Dot" }, returnTypes = {
		Dot.class }, parameterLabels = { "DPN-net", "Initial marking", "Final marking" }, userAccessible = true)
public class DPNGraphvizConverterPlugin {

	public enum VariableDisplayMode {
		HIDDEN, 
		/**
		 * Visible variables without influence on the layout 
		 */
		NO_LAYOUT, 
		/**
		 * Visible variables with influence on the layout
		 */
		AUTO_LAYOUT,
		NO_LAYOUT_NO_READS, 
		AUTO_LAYOUT_NO_READS		
	}

	public enum PlaceDisplayMode {
		BASIC, DETAILED
	}

	public enum GuardDisplayMode {
		HIDDEN, TRANSITIONS, EDGES
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", uiLabel = "Convert DPN-net to DOT")
	@PluginVariant(variantLabel = "Convert DPN-net to DOT", requiredParameterLabels = { 0 })
	public Dot dataPetriNetToDotCombined(UIPluginContext context, DataPetriNet petrinet) throws UserCancelledException {

		GraphDirection direction = GraphDirection.leftRight;
		GuardDisplayMode guardDisplayMode = GuardDisplayMode.EDGES;
		VariableDisplayMode variableDisplayMode = VariableDisplayMode.NO_LAYOUT;
		PlaceDisplayMode placeDisplayMode = PlaceDisplayMode.BASIC;

		direction = ProMUIHelper.queryForObject(context, "Layout orientation", GraphDirection.values());
		guardDisplayMode = ProMUIHelper.queryForObject(context, "Guard display mode", GuardDisplayMode.values());
		variableDisplayMode = ProMUIHelper.queryForObject(context, "Variable display mode",
				VariableDisplayMode.values());
		placeDisplayMode = ProMUIHelper.queryForObject(context, "Place display mode", PlaceDisplayMode.values());

		return DPNGraphvizConverter.convertDPN(petrinet, MarkingsHelper.getInitialMarkingOrEmpty(context, petrinet),
				MarkingsHelper.getFinalMarkingsOrEmpty(context, petrinet), guardDisplayMode, variableDisplayMode,
				placeDisplayMode, direction).getDot();
	}

}