package org.processmining.datapetrinets.visualization.graphviz;

import org.apache.commons.lang3.StringEscapeUtils;
import org.processmining.plugins.graphviz.dot.DotNode;

import com.google.common.collect.ImmutableMap;

class PlaceNode extends DotNode {

	private static final int MAX_LINE_LENGTH = 8;
	private static final int MAX_PLACE_LABEL = 27;

	public PlaceNode() {
		super("", ImmutableMap.of("shape", "circle"));		
	}

	public PlaceNode(String label) {
		super(genLabel(label), ImmutableMap.of("shape", "circle"));
	}

	private static String genLabel(String label) {
		if (label.isEmpty()) {
			return label;
		} else {
			String escapeXml = StringEscapeUtils.escapeXml(label);
			if (escapeXml.length() > MAX_PLACE_LABEL) {
				escapeXml = org.apache.commons.lang3.StringUtils.abbreviate(escapeXml, MAX_PLACE_LABEL);
			}
			if (escapeXml.length() > MAX_LINE_LENGTH) {
				escapeXml = VisualizationUtils.wrapLineWithSeparator(escapeXml, "<BR/>", MAX_LINE_LENGTH);
			}
			return "<<B>" + escapeXml + "</B>>";
		}
	}

	public String toString() {
		String result = "\"" + getId() + "\" [label=" + labelToString() + ", id=\"" + getId() + "\"";
		for (String key : getOptionKeySet()) {
			result += "," + key + "=\"" + getOption(key) + "\"";			
		}
		if (getLabel() == null || getLabel().isEmpty()) {
			result += ",width=0.4";
			result += ",fixedsize=true";
		}
		return result + "];";
	}

}