package org.processmining.models.graphbased.directed.petrinetwithdata.newImpl;

import java.awt.Color;
import java.io.Serializable;
import java.util.StringTokenizer;

import javax.swing.SwingConstants;

import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.NodeID;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.shapes.Ellipse;

@SuppressWarnings("unchecked")
public class DataElement extends PetrinetNode implements Serializable {

	private static final long serialVersionUID = -8923126527923571380L;
	
	private final String varName;
	private final Class<?> type;
	private final Comparable<?> minValue;
	private final Comparable<?> maxValue;
	private final NodeID id = new NodeID();

	@SuppressWarnings("rawtypes")
	public DataElement(String varName, Class type, Comparable minValue, Comparable maxValue, PetriNetWithData petrinet) {
		super(petrinet, null, varName);
		if (minValue != null && !minValue.getClass().equals(type))
			throw new IllegalArgumentException("The minimum value is incompatible with " + type);
		if (maxValue != null && !maxValue.getClass().equals(type))
			throw new IllegalArgumentException("The maximum value is incompatible with " + type);
		if (maxValue != null && minValue != null && maxValue.compareTo(minValue) < 0)
			throw new IllegalArgumentException(maxValue + "is smaller than " + minValue);
		this.varName = varName;
		this.minValue = minValue;
		this.maxValue = maxValue;
		this.type = type;
		//getAttributeMap().put(AttributeMap.SIZE, new Dimension(stdWidth, stdHeight));
		getAttributeMap().put(AttributeMap.RESIZABLE, true);
		getAttributeMap().put(AttributeMap.SHOWLABEL, true);
		getAttributeMap().put(AttributeMap.FILLCOLOR, Color.ORANGE);
		getAttributeMap().put(AttributeMap.SHAPE, new Ellipse());

		getAttributeMap().put(AttributeMap.TOOLTIP, getToolTip());
		getAttributeMap().put(AttributeMap.LABELVERTICALALIGNMENT, SwingConstants.CENTER);
		getAttributeMap().put(AttributeMap.LABEL,
				"<html><center><b>" + encodeTextAsHTML(varName) + "</b></center></html>");
	}

	private String getToolTip() {
		String tooltip = "<html><center><table><tr><td><b>Variable Name:</b></td><td>" + varName + "</td></tr>"
				+ "<tr><td><b>Attribute type:</b></td><td>" + encodeTextAsHTML(type.toString()) + "</td></tr>";
		if (minValue != null)
			tooltip += "<tr><td><b>Minimum Value:</b></td><td>" + minValue + "</td></tr>";
		if (maxValue != null)
			tooltip += "<tr><td><b>Maximum Value:</b></td><td>" + maxValue + "</td></tr>";
		tooltip += "</table></html>";
		return tooltip;
	}

	private static String encodeTextAsHTML(String pText) {
		StringTokenizer tokenizer = new StringTokenizer(pText, "&<>\"", true);
		int tokenCount = tokenizer.countTokens();

		/* no encoding's needed */
		if (tokenCount == 1)
			return pText;

		/*
		 * text.length + (tokenCount * 6) gives buffer large enough so no
		 * addition memory would be needed and no costly copy operations would
		 * occur
		 */
		StringBuffer buffer = new StringBuffer(pText.length() + tokenCount * 6);
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			if (token.length() == 1) {
				switch (token.charAt(0)) {
					case '&' :
						buffer.append("&amp;");
						break;
					case '<' :
						buffer.append("&lt;");
						break;
					case '>' :
						buffer.append("&gt;");
						break;
					case '"' :
						buffer.append("&quot;");
						break;
					default :
						buffer.append(token);
				}
			} else {
				buffer.append(token);
			}
		}
		return buffer.toString();
	}

	public String toString() {
		return getVarName();
	}

	public String getVarName() {
		return varName;
	}

	@SuppressWarnings("rawtypes")
	public Comparable getMinValue() {
		return minValue;
	}

	@SuppressWarnings("rawtypes")
	public Comparable getMaxValue() {
		return maxValue;
	}

	@SuppressWarnings("rawtypes")
	public Class getType() {
		return type;
	}

	@Override
	public int hashCode() {
		return varName.hashCode();
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		DataElement other = (DataElement) obj;
		return varName.equals(other.varName);
	}

	public NodeID getId() {
		return id;
	}

}
